# include <pso.h>
# include <math.h>
# include "gradientwithls.h"
# define SIMCHECK
# define LOCALSEARCH

static double dmin(double a,double b)
{	
	return a<b?a:b;
}


static double dmax(double a,double b)
{
	return a>b?a:b;
}

static double getNorm(Data x)
{
	double s=0.0;
	for(int i=0;i<x.size();i++) s+=x[i]*x[i];
	return sqrt(s);
}

static double getDistance(Data x1,Data x2)
{
	double s=0.0;
	for(int i=0;i<x1.size();i++) s+=(x1[i]-x2[i])*(x1[i]-x2[i]);
	return s;
}

static double maxElement(Data x)
{
	double m=-1e+100;
	for(int i=0;i<x.size();i++) if(x[i]>m) m=x[i];
	return m;
}


static double minElement(Data x)
{
	double m=1e+100;
	for(int i=0;i<x.size();i++) if(x[i]<m) m=x[i];
	return m;
}

static void localSearch(Problem *p,Data &x,double &y)
{
    GradientWithLS gd(p);
    gd.setPoint(x,y);
    gd.setLineSearch("armijo");
    gd.Solve();
    x= gd.getBestx();
    y = gd.getBesty();
}


Pso::Pso(Problem *p,int count,int type)
{
	pso_type = type;
	problem = p;
	xbest.resize(p->getDimension());
	ybest   = 1e+100;
	generation = 0;
    particle.resize(count);
	fitness.resize(count);
    bestparticle.resize(count);
	bestFitness.resize(count);	
	oldBestFitness.resize(count);	
	vel.resize(count);
	lmargin.resize(problem->getDimension());
	rmargin.resize(problem->getDimension());
    lmargin= problem->getLeftMargin();
    rmargin = problem->getRightMargin();
	center.resize(problem->getDimension());
	for(int i=0;i<count;i++)
	{
        particle[i].resize(problem->getDimension());
		vel[i].resize(problem->getDimension());
		for(int j=0;j<vel[i].size();j++) vel[i][j]=0.0;
        bestparticle[i].resize(problem->getDimension());
       particle[i]=problem->getSample();

        bestparticle[i]=particle[i];
        fitness[i]=problem->statFunmin(particle[i]);
		bestFitness[i]=fitness[i];
		oldBestFitness[i]=bestFitness[i];
		if(fitness[i]<ybest)
		{
			ybest = fitness[i];
            xbest = particle[i];
			bestIndex = i;
		}
	}
}

void	Pso::calcFitnessArray()
{
    for(int i=0;i<particle.size();i++)
	{
			oldBestFitness[i]=bestFitness[i];
#ifdef SIMCHECK
        double dist=getDistance(particle[i],bestparticle[i]);
		if(dist<1e-5)
		{
            particle[i]=bestparticle[i];
			fitness[i]=bestFitness[i];
			continue;
		}
#endif
		
#ifdef LOCALSEARCH
	
        if(rand()%particle.size()==0)
            localSearch(problem,particle[i],fitness[i]);
		else      
	
#endif
        fitness[i]=problem->statFunmin(particle[i]);

		if(fitness[i]<bestFitness[i])
		{
			bestFitness[i]=fitness[i];
            bestparticle[i]=particle[i];
		}
	}
	updateBest();
}

void	Pso::updateBest()
{
	for(int i=0;i<fitness.size();i++)
	{
		if(fitness[i]<ybest)
		{
			ybest = fitness[i];
            xbest = particle[i];
			bestIndex=i;
		}
	}
	if(pso_type==CENTERPSO)
	{
        double centerValue = problem->statFunmin(center);
		if(centerValue<ybest)
		{
			ybest = centerValue;
			xbest = center;
		}
	}
}

void    Pso::setGenerations(int g)
{
        if(g>0) generations = g;
}
int     Pso::getGenerations() const
{
    return generations;
}

double	Pso::getInertia(int index)
{
	if(pso_type==DPSO)
	{
		double u=1.0002;
		double w=0.3;
		return w * pow (u,-1.0 * generations );
		/*
		double omega=0.0;
		double meanFitness=0.0;
		int    countFitness=0;
		double minF=1e+100;
		for(int i=0;i<fitness.size();i++)
		{
			countFitness++;
			meanFitness+=fitness[i];
			if(fitness[i]<minF) minF=fitness[i];
		}
		meanFitness/=countFitness;

		double h1=dmin(bestFitness[index],oldBestFitness[index]);
		double h2=dmax(bestFitness[index],oldBestFitness[index]);
		double ht=fabs(h1/h2);
		double s=fabs(dmin(minF,meanFitness)/dmax(minF,meanFitness));

		double omegaini=1.0;
		double alpha=0.4;
		double beta=0.8;
		omega=omegaini-alpha*(1.0-ht)+beta*s;
		return omega;
		*/
	}
	return 0.9-0.5*generation*1.0/generations;
}

void	Pso::boundVelocity(double &v,int index)
{
	const double gamma=0.5; //proposed by Ali
	double umax=gamma * (rmargin[index]-lmargin[index]);
	double umin=-umax;
	if(v<umin) v=umin;
	if(v>umax) v=umax;
}

void	Pso::updateCenter()
{
	for(int j=0;j<problem->getDimension();j++) center[j]=0.0;
    for(int i=0;i<particle.size();i++)
	{
		for(int j=0;j<problem->getDimension();j++)
		{
            center[j]+=particle[i][j];
		}
	}
	for(int j=0;j<problem->getDimension();j++) 
	{
        center[j]/=particle.size();
		if(center[j]<lmargin[j]) center[j]=lmargin[j];
		if(center[j]>rmargin[j]) center[j]=rmargin[j];
	}
}

void	Pso::nextGeneration()
{
	++generation;
	double psi1=1.0,psi2=1.0,psi3=1.0;
	Data xp;xp.resize(problem->getDimension());
	double total_mean=0.0;
	double r1,r2,v1,v2,accel,accel_sign,trial_accel;
    for(int i=0;i<particle.size();i++)
	{
		if(pso_type==DPSO)
		{
			for(int j=0;j<problem->getDimension();j++)
			{
				double r1=drand48();
				double r2=drand48();
                vel[i][j]=getInertia(i)*vel[i][j]+psi1*r1*(xbest[j]-particle[i][j])+
                    psi2*r2*(bestparticle[i][j]-particle[i][j]);
				boundVelocity(vel[i][j],j);
                particle[i][j]=particle[i][j]+vel[i][j];
                if(particle[i][j]<lmargin[j])
				{	
                    particle[i][j]=lmargin[j];
				}
                if(particle[i][j]>rmargin[j])
				{
                    particle[i][j]=rmargin[j];
				}
			}
		}
		else
		for(int j=0;j<problem->getDimension();j++)
		{
			r1=drand48();
			r2=drand48();
			switch(pso_type)
			{
				case LDWPSO:
				case CENTERPSO:
                    vel[i][j]=getInertia(i)*vel[i][j]+psi1*r1*(xbest[j]-particle[i][j])+
                        psi2*r2*(bestparticle[i][j]-particle[i][j]);
					boundVelocity(vel[i][j],j);
                    particle[i][j]=particle[i][j]+vel[i][j];
					break;
				case PPSO:
                    vel[i][j]=0.729*(vel[i][j]+2.05*r1*(xbest[j]-particle[i][j])+
                        2.05*r2*(bestparticle[i][j]-particle[i][j]));
					boundVelocity(vel[i][j],j);
                    particle[i][j]=particle[i][j]+vel[i][j];
					break;
				case GPSO:
                    v1=psi1 * r1 *(xbest[j]-particle[i][j]);
                    v2=psi2 * r2 *(bestparticle[i][j]-particle[i][j]);

					accel=v1+v2;
                    particle[i][j]=particle[i][j]+accel;
					break;
			}
            if(particle[i][j]<lmargin[j])
			{	
                particle[i][j]=lmargin[j];
			}
            if(particle[i][j]>rmargin[j])
			{
                particle[i][j]=rmargin[j];
			}
		}
	}
	if(pso_type==CENTERPSO) updateCenter();
}

void	Pso::Solve(int iprint)
{
	double x1=0.0;
	double x2=0.0;
	double oldbest = 1e+100;
	double stopat;
	generation=1;
    for(int i=2;i<=generations+1;i++)
	{
		nextGeneration();
		calcFitnessArray();
		
		if(iprint) printf("ITERATION=%4d\tYBEST=%20.8lg\t",i,ybest);
		double v=fabs(ybest);
		x1+=v;
		x2+=v*v;
		double variance=x2/i-(x1/i)*(x1/i);
		if(variance<0) variance=-variance;
		if(fabs(ybest-oldbest)>1e-4)
		{
			oldbest=ybest;
			stopat=variance/2.0;
		}
		if(stopat<1e-7) stopat=variance/2.0;
		if(iprint) 
		printf("VARIANCE=%10.4lg\tSTOPAT=%10.4lg\n",variance,stopat);
		if(variance<stopat && i>=5) break;
	}
	localSearch(problem,xbest,ybest);
}

int	Pso::getGeneration()
{
	return generation;
}

int	Pso::getparticleCount()
{
    return particle.size();
}

int	Pso::getDimension()
{
	return problem->getDimension();
}

void	Pso::getBest(Data &x,double &y)
{
	x = xbest;
	y = ybest;
}

Pso::~Pso()
{
}
